#!/usr/bin/python2.7
""" Frack Bar application
This application is intended to track all bar sales made at the
Fryslan Hackerspace FRACK.

Recent changes:
v0.1
- Initial version

v0.2
- Sanitized input
- Several bug fixes and code cleanup
- first production version!

v0.3:
- Changed the sequence to first ask for payment method
- Changed product input to allow for multiple products
- Input policy is changed, it now only accepts actual valid input
- ADMIN section added with CSV Export """
__author__ = 'Rudi Daemen <info@kratjebierhosting.nl>'
__version__ = '0.3'

# Required modules
import sqlite3
import datetime
import os
import time
import sys
import csv

class FrackDb(object):
  """ Class to handle all Database access for the application """
  def __init__(self, dbfile=':memory'):
    self.connection = sqlite3.connect(dbfile)

  def GetProduct(self, barcode):
    """ method to get only the required product info from products """
    with self.connection as cursor:
      return cursor.execute("""select `ID`,`name`,`memberprice`,`visitorprice`
                        from `product`
                        where `barcode` = ? ORDER BY `ID` DESC LIMIT 1""",
                        (barcode,)).fetchone()

  def GetCard(self, barcode):
    """ method to get only the required creditcard info from creditcard """
    with self.connection as cursor:
      return cursor.execute("""select `ID`,`member`,`credit`
                        from `creditcard`
                        where `barcode` = ? ORDER BY `ID` DESC LIMIT 1""",
                        (barcode,)).fetchone()

  def GetSales(self, card, member):
    """ method to get the sum of all sales associated with creditcard the ID.
    This is based on whether the card is registered for a member or a
    visitor. It does a multi-table select and sums up the result.  """
    with self.connection as cursor:
      if member:
        return cursor.execute("""
            SELECT SUM(`product`.`memberprice`)
            FROM `product`
            JOIN `sales` ON `sales`.`product` = `product`.`ID`
            WHERE `sales`.`card` = ? """, (card,)).fetchone()
      else:
        return cursor.execute("""
            SELECT SUM(`product`.`visitorprice`)
            FROM `product`
            JOIN `sales` ON `sales`.`product` = `product`.`ID`
            WHERE `sales`.`card` = ? """, (card,)).fetchone()

  def SetSale(self, product, card=None, member=False):
    """ This method is the actual data writer. It by default asumes the sale
    was done by cash method and a non-member. Returns a string to indicate
    success or failure. """
    with self.connection as cursor:
      cursor.execute("""
        insert into sales (`datetime`, `product`, `card`, `member`)
        values (?, ?, ?, ?)""",
        (datetime.datetime.now().strftime('%F %T'), product, card, member))
    return True

  def ProdSales(self, date=''):
    """ This method gets all sales data for each product """
    date = date + '%'
    sales = []
    with self.connection as cursor:
      result = cursor.execute("SELECT `ID` from `product`").fetchall()
    for items in result:
      with self.connection as cursor:
        sales.append(cursor.execute("""
          SELECT `product`.`ID`, `product`.`name`, sum(`sales`.`member` = '1')
             * `product`.`memberprice` AS MemTurn, sum(`sales`.`member` = '0')
             * `product`.`visitorprice` AS VisTurn, count(`sales`.`product`)
          FROM `sales`
          JOIN `product` ON `sales`.`product` =  `product`.`ID`
          WHERE `product`.`ID` = ? and `sales`.`datetime` LIKE ? """,
          (items[0], date)).fetchone())
    return sales

  def CardSales(self, date=''):
    """ This method returns all creditcards, creation date, credit and used
    credit """
    date = date + '%'
    cards = []
    with self.connection as cursor:
      result = cursor.execute("""SELECT `ID`,`member` FROM `creditcard`
          WHERE `datetime` LIKE ? """, (date,)).fetchall()
    for items in result:
      with self.connection as cursor:
        cards.append(cursor.execute("""
          SELECT `ID`, `datetime`, `barcode`, `member`, `credit`
          FROM `creditcard` WHERE `ID` = ? """,
          (items[0],)).fetchone() + self.GetSales(items[0], items[1]))
    return cards

  def GetAllTable(self, table):
    """ This method returns a dictionary of table names and values from given
    table. """
    with self.connection as cursor:
      data = cursor.execute("SELECT * FROM `%s`" % table)
    headers = [field[0] for field in data.description]
    return [dict(zip(headers, row)) for row in data.fetchall()], headers


def AdminCards(conn, date=''):
  """ This function returns all cards within given date and shows how much
  credit the card still has left based on sales. """
  cards = conn.CardSales(date)
  print (u"\033[1m\nCard\tCreation\t\t\t\tFor\tInitial\tRemaining"
      u"\033[1m\nID\tdate\t\t\tBarcode\t\tMember?\tcredit\tcredit\033[0;0m"
      u"\n-------+-----------------------+---------------+-------+-------"
      u"+--------")
  for items in cards:
    if items[5]:
      remaining = items[4] - items[5]
    else:
      remaining = items[4]
    if items[3] == 1:
      member = 'YES'
    else:
      member = 'NO'
    print (u"%g\t%s\t%s\t%s\t\u20AC%.2f\t\u20AC%.2f"
        % (items[0], items[1], items[2], member, items[4], remaining))


def AdminSales(conn, date=''):
  """ This function returns all sales and turnover within given date and
  shows the total turnover per group and summed up. """
  sales = conn.ProdSales(date)
  print u"\033[1m\nProduct\tProduct\t\t\tVisitor\tMember\tNumber"
  print u"ID\tName\t\t\tincome\tincome\tsold\033[0;0m"
  print u"-------+-----------------------+-------+-------+------"
  visturn = 0
  memturn = 0
  for items in sales:
    if items[0]:
      visturn = visturn + items[3]
      memturn = memturn + items[2]
      prodname = items[1].ljust(24)
      print (u"%g\t%s\u20AC%.2f\t\u20AC%.2f\t%g"
          % (items[0], prodname, items[3], items[2], items[4]))
  totturn = visturn + memturn
  print u"-------+-----------------------+-------+-------+------"
  print u"Total Visitor income:\t\033[1m\u20AC%.2f\033[0;0m" % visturn
  print u"Total Member income:\t\033[1m\u20AC%.2f\033[0;0m" % memturn
  print u"\033[1mTOTAL TURNOVER:\t\t\u20AC%.2f\033[0;0m" % totturn

def BarBanner():
  """ The header of the application, it clears the screen to clean it up for
  the next transaction to be done and then prints the header again """
  os.system("clear")
  print ' _____               _      ____'
  print '|  ___| __ __ _  ___| | __ | __ )  __ _ _ __'
  print '| |_ | \'__/ _` |/ __| |/ / |  _ \ / _` | \'__|'
  print '|  _|| | | (_| | (__|   <  | |_) | (_| | |'
  print '|_|  |_|  \__,_|\___|_|\_\ |____/ \__,_|_|'
  print ' Frack Consumption Registration System %s' % __version__
  print ''


def ScanItem(conn):
  """ Scan function, this is intended handle some special codes which allows
  the user to start an Admin process or quit the program or restart because he
  made a mistake. """
  data = raw_input().decode('utf-8')
  special_codes = ('QUIT', 'RESET', 'ADMIN')
  if data in special_codes:
    print '\nINTERRUPT: Special code detected!'
    if data == 'QUIT':
      print "\nClosing application..."
      sys.exit(0)
    elif data == 'RESET':
      print 'Resetting application ...'
      time.sleep(3)
    elif data == 'ADMIN':
      print 'Entering administration mode ...'
      KassaAdmin(conn)
  else:
    # Do standard input stuff.
    return data


def KassaAdmin(conn):
  """ This will be the administrative function in the program """
  done = False
  while not done:
    BarBanner()
    print (u"Administrative menu\n\n"
        u"\033[1m10\033[0;0m: Sales overview\n"
        u" \033[1m11\033[0;0m: Last years sales\t"
        u" \033[1m12\033[0;0m: This years sales\n"
        u" \033[1m13\033[0;0m: Last months sales\t"
        u" \033[1m14\033[0;0m: This months sales\n"
        u"\n"
        u"\033[1m20\033[0;0m: CreditCards overview\n"
        u" \033[1m21\033[0;0m: Last years cards\t"
        u" \033[1m22\033[0;0m: This years cards\n"
        u"\n"
        u"\033[1m30\033[0;0m: Export all to CSV\n"
        u" \033[1m31\033[0;0m: Export products\t"
        u" \033[1m32\033[0;0m: Export creditcards\n"
        u" \033[1m33\033[0;0m: Export sales\t"
        u"\n"
        u"\n\033[1mDONE\033[0;0m: Exit Admin menu\n")
    scan = ''
    date = ''
    what = ''
    scan = raw_input(u"Please scan your option: ").decode('utf8')

    if scan == 'DONE':
      done = True
      continue
    elif scan.startswith('1'):
      if scan == '11':
        date = (datetime.datetime.now() - datetime.timedelta(365)
            ).strftime('%Y')
      elif scan == '12':
        date = datetime.datetime.now().strftime('%Y')
      elif scan == '13':
        date = (datetime.datetime.now() - datetime.timedelta(365/12)
            ).strftime('%Y-%m')
      elif scan == '14':
        date = datetime.datetime.now().strftime('%Y-%m')
      print u"Selecting all sales matching date: '\033[1m%s*\033[0;0m'" % date
      AdminSales(conn, date)
    elif scan.startswith('2'):
      if scan == '21':
        date = (datetime.datetime.now() - datetime.timedelta(365)
            ).strftime('%Y')
      elif scan == '22':
        date = datetime.datetime.now().strftime('%Y')
      print u"Selecting all cards matching date: '\033[1m%s*\033[0;0m'" % date
      AdminCards(conn, date)
    elif scan.startswith('3'):
      if scan == '31':
        what = 'prod'
      elif scan == '32':
        what = 'card'
      elif scan == '33':
        what = 'sale'
      AdminCsv(conn, what)
    else:
      print "Option not (yet) implemented."
    raw_input('\nScan anything to return to Admin menu...').decode('utf8')
  return True

def AdminCsv(conn, what=''):
  """ Export to CSV function. """
  filename = ''
  timestamp = datetime.datetime.now().strftime('%F_%H%M')
  if what == 'prod':
    tables = ('product',)
  elif what == 'card':
    tables = ('creditcard',)
  elif what == 'sale':
    tables = ('sales',)
  else:
    tables = ('product', 'sales', 'creditcard')
  for item in tables:
    data = conn.GetAllTable(item)
    filename = timestamp + '_' + item + '.csv'
    with file(filename, 'w') as source:
      writer = csv.DictWriter(source, data[1])
      writer.writeheader()
      writer.writerows(data[0])
    print (u"Dumped table '\033[1m%s\033[0;0m' to file '\033[1m%s\033[0;0m'."
        % (item, filename))


def ScanProd(conn, ismember):
  """ This is the multi-select function for product scanning. The function
  requires that at least one product is scanned and it returns a list of
  products """
  barcode = ''
  products = []
  num = 0
  while not barcode == 'DONE' or num == 0:
    barcode = raw_input("Please scan your product(s) or \033[1mDONE\033[0;0m "
        "when you are done: ").decode('utf-8')
    if not barcode == 'DONE':
      prodinfo = conn.GetProduct(barcode)
      if not prodinfo:
        print "ERROR: Invalid product code scanned."
      else:
        num = num + 1
        products.append(prodinfo)
    if num == 0:
      print "ERROR: You need to scan at least one product!"
  print "\nYou have scanned the following products: "
  total = 0
  for item in products:
    price = item[3]
    if ismember == '1':
      price = item[2]
    total = total + price
    print (u"- \033[1m%s\033[0;0m; price: \033[1m\u20AC %.2f\033[0;0m."
        % (item[1], price))
  print u"This sums up to a total of \033[1m\u20AC %.2f\033[0;0m.\n" % total
  return products


def CashMode(conn):
  """ Cash-only transactions are handled in here """
  print u"\nYou selected \033[1mCASH\033[0;0m mode.\n"
  scan = ''
  while not scan == 'MEMBER' and not scan == 'VISITOR':
    scan = raw_input(u"Please scan if you are a \033[1mMEMBER\033[0;0m"
        u" or a \033[1mVISITOR\033[0;0m: ").decode('utf-8')
    if not scan == 'MEMBER' and not scan == 'VISITOR':
      print "ERROR: Invalid input. Either scan MEMBER or VISITOR."
  if scan == 'MEMBER':
    ismember = '1'
  else:
    ismember = '0'
  
  products = ScanProd(conn, ismember)
  result = AccSale()
  if result:
    for item in products:
      conn.SetSale(item[0], '', ismember)
    return True
  else:
    return False


def CcMode(card, conn):
  """ Creditcard transactions are handled in here """
  cardsales = conn.GetSales(card[0], card[1])
  if cardsales[0]:
    credit = card[2] - cardsales[0]
  else:
    credit = card[2]

  print (u"\nYou selected CreditCard mode. You card has \033[1m\u20AC %.2f"
      u"\033[0;0m remaining.\n" % credit)

  products = ScanProd(conn, card[1])
  total = 0
  for item in products:
    if card[1] == '1':
      total = total + item[2]
    else:
      total = total + item[3]

  cardsales = conn.GetSales(card[0], card[1])
  if cardsales[0]:
    credit = card[2] - cardsales[0]
  else:
    credit = card[2]

  if total > credit:
    print (u"ERROR: Not enough credit (\033[1m\u20AC %.2f\033[0;0m) on your "
        "card!" % credit)
    return False
  else:
    result = AccSale()

  if result:
    for item in products:
      conn.SetSale(item[0], card[0], card[1])
    return True
  else:
    return False


def AccSale():
  """ Simple ACCEPT/CANCEL function. Returns True or False. """
  sold = ''
  while not sold == 'ACCEPT' and not sold == 'CANCEL':
    sold = raw_input(u"Please scan \033[1mACCEPT\033[0;0m to confirm or "
        u"\033[1mCANCEL\033[0;0m to abort:\n\n ").decode('utf-8')
    if not sold == 'ACCEPT' and not sold == 'CANCEL':
      print "ERROR: Invalid input."
  if sold == 'ACCEPT':
    return True
  else:
    return False


def Kassa():
  """ The main application, this handles all input and sends it to the DB
  handler class FrackDb. """
  conn = FrackDb('frack_consump.sqlite')
  while True:
    BarBanner()

    payment = False
    cardinfo = False
    while not payment == 'CASH' and not cardinfo:
      payment = raw_input(u"Please scan \033[1mCASH\033[0;0m or your \033[1m"
          u"CreditCard\033[0;0m: ").decode('utf-8')
      if payment == 'ADMIN':
        KassaAdmin(conn)
        BarBanner()
        continue

      cardinfo = conn.GetCard(payment)
      if not cardinfo and not payment == 'CASH':
        print u"ERROR: Invalid Card scanned."

    if payment == 'CASH':
      result = CashMode(conn)
    else:
      result = CcMode(cardinfo, conn)

    if result:
      print "\nTransaction completed successfully. Resetting in 5 seconds..."
    else:
      print "\nTransaction aborted. Resetting in 5 seconds..."

    print "\n\nThank you, come again! (c) Apu Nahasapeemapetilon, The Simpsons."
    time.sleep(5)


if __name__ == '__main__':
  Kassa()
